/* jcifs smb client library in Java
 * Copyright (C) 2000  "Michael B. Allen" <jcifs at samba dot org>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jcifs;

import jcifs.util.Log;
import java.util.Properties;
import java.util.StringTokenizer;
import java.io.InputStream;
import java.io.PrintStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.net.InetAddress;

/**
 * This class mananges configuration information specific to the jCIFS library. There are several ways to set jCIFS properties. See the <a href="../overview-summary.html#scp">overview page of the API documentation</a> for details.
 */

public class Config {

/**
 * This static method registers the SMB URL protocol handler which is required to use SMB URLs with the <tt>java.net.URL</tt> class. If this method is not called before attempting to create an SMB URL with the URL class the following exception will occur:
<blockquote><pre>
Exception MalformedURLException: unknown protocol: smb
      at java.net.URL.<init>(URL.java:480)
      at java.net.URL.<init>(URL.java:376)
      at java.net.URL.<init>(URL.java:330)
      at jcifs.smb.SmbFile.<init>(SmbFile.java:355)
      ...
</pre><blockquote>
 */

	public static void registerSmbURLHandler() {
		String ver, pkgs;

		ver = System.getProperty( "java.version" );
		if( ver.startsWith( "1.1." ) || ver.startsWith( "1.2." )) {
		     throw new RuntimeException( "jcifs-0.7.0b4+ requires Java 1.3 or above. You are running " + ver );
		}
		pkgs = System.getProperty( "java.protocol.handler.pkgs" );
		if( pkgs == null ) {
			pkgs = "jcifs";
		} else {
			pkgs += "|jcifs";
		}
		System.setProperty( "java.protocol.handler.pkgs", pkgs );
	}

	static {
		String filename, mask;
		FileInputStream in = null;

		try {
			filename = System.getProperty( "jcifs.properties" );
			if( filename != null && filename.length() > 1 ) {
				in = new FileInputStream( filename );
			}
			jcifs.util.Config.load( in );
		} catch( IOException ioe ) {
			ioe.printStackTrace();
		}

		if(( mask = jcifs.util.Config.getProperty( "jcifs.util.log" )) != null ) {

			StringTokenizer st = new StringTokenizer( mask, "," );
			while( st.hasMoreTokens() ) {
				String s = st.nextToken().toUpperCase();
				if( s.startsWith( "ALL" )) {
					Log.addMask( Log.ALL );
				} else if( s.startsWith( "EXC" )) {
					Log.addMask( Log.EXCEPTIONS );
				} else if( s.startsWith( "WAR" )) {
					Log.addMask( Log.WARNINGS );
				} else if( s.startsWith( "DEB" )) {
					Log.addMask( Log.DEBUGGING );
				} else if( s.startsWith( "HEX" )) {
					Log.addMask( Log.HEX_DUMPS );
				} else if( s.startsWith( "NON" )) {
					Log.setMask( Log.NONE );
					break;
				}
			}
		}
		jcifs.util.Config.printProperties( "properties loaded" );
	}

	public static void load( InputStream in ) throws IOException {
		jcifs.util.Config.load( in );
	}
	public static void list( PrintStream out ) throws IOException {
		jcifs.util.Config.list( out );
	}
	public static String setProperty( String key, String value ) {
		return jcifs.util.Config.setProperty( key, value );
	}
	public static Object get( String key ) {
		return jcifs.util.Config.get( key );
	}
	public static String getProperty( String key, String def ) {
		return jcifs.util.Config.getProperty( key, def );
	}
	public static String getProperty( String key ) {
		return jcifs.util.Config.getProperty( key );
	}
	public static int getInt( String key, int def ) {
		return jcifs.util.Config.getInt( key, def );
	}
	public static int getInt( String key ) {
		return jcifs.util.Config.getInt( key );
	}
	public static InetAddress getInetAddress( String key, InetAddress def ) {
		return jcifs.util.Config.getInetAddress( key, def );
	}
	public static boolean getBoolean( String key, boolean def ) {
		return jcifs.util.Config.getBoolean( key, def );
	}
}

