/* jcifs smb client library in Java
 * Copyright (C) 2000  "Michael B. Allen" <jcifs at samba dot org>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jcifs.netbios;

import java.net.InetAddress;
import java.net.DatagramSocket;
import java.net.DatagramPacket;
import java.net.UnknownHostException;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.util.Hashtable;
import java.util.StringTokenizer;
import jcifs.Config;

class NameServiceClient implements Runnable {

    static final int DEFAULT_SO_TIMEOUT = 5000;
    static final int DEFAULT_RCV_BUF_SIZE = 576;
    static final int DEFAULT_SND_BUF_SIZE = 576;
	static final int NAME_SERVICE_UDP_PORT = 137;
	static final int DEFAULT_RETRY_COUNT = 2;
	static final int DEFAULT_RETRY_TIMEOUT = 3000;

	static final int RESOLVER_LMHOSTS = 1;
	static final int RESOLVER_BCAST   = 2;
	static final int RESOLVER_WINS    = 3;

    InetAddress laddr, baddr, nbns;
    int port, soTimeout, retryCount, retryTimeout, closeTimeout;
    int snd_buf_size, rcv_buf_size;
    byte[] snd_buf, rcv_buf;
    DatagramSocket socket;
    DatagramPacket in, out;
	Hashtable responseTable;
    Thread thread;
	int nextNameTrnId;
	Object socketLock;
	int[] resolveOrder;

	NameServiceClient() {
		this( Config.getInt( "jcifs.netbios.lport", 0 ),
			Config.getInetAddress( "jcifs.netbios.laddr", null ));
	}
    NameServiceClient( int port, InetAddress laddr ) {
        this.port = port;
        this.laddr = laddr;

		try {
			baddr = Config.getInetAddress( "jcifs.netbios.baddr",
						InetAddress.getByName( "255.255.255.255" ));
		} catch( UnknownHostException uhe ) {
		}

		nbns = Config.getInetAddress( "jcifs.netbios.wins", null );
        soTimeout = Config.getInt( "jcifs.netbios.soTimeout", DEFAULT_SO_TIMEOUT );
		retryCount = Config.getInt( "jcifs.netbios.retryCount", DEFAULT_RETRY_COUNT );
		retryTimeout = Config.getInt( "jcifs.netbios.retryTimeout", DEFAULT_RETRY_TIMEOUT);
        rcv_buf_size = Config.getInt( "jcifs.netbios.rcv_buf_size", DEFAULT_RCV_BUF_SIZE );
        snd_buf_size = Config.getInt( "jcifs.netbios.snd_buf_size", DEFAULT_SND_BUF_SIZE );
        rcv_buf = new byte[rcv_buf_size];
        snd_buf = new byte[snd_buf_size];
        in  = new DatagramPacket( rcv_buf, rcv_buf_size );
        out = new DatagramPacket( snd_buf, snd_buf_size, baddr, NAME_SERVICE_UDP_PORT );
		responseTable = new Hashtable();
		nextNameTrnId = 0;
		socketLock = new Object();

		String ro = Config.getProperty( "jcifs.resolveOrder" );
		if( ro == null || ro.length() == 0 ) {

			/* No resolveOrder has been specified, use the
			 * default which is LMHOSTS,WINS,BCAST,DNS or just
			 * LMHOSTS,BCAST,DNS if jcifs.netbios.wins has not
			 * been specified.
			 */

			if( nbns == null ) {
				resolveOrder = new int[2];
				resolveOrder[0] = RESOLVER_LMHOSTS;
				resolveOrder[1] = RESOLVER_BCAST;
			} else {
				resolveOrder = new int[3];
				resolveOrder[0] = RESOLVER_LMHOSTS;
				resolveOrder[1] = RESOLVER_WINS;
				resolveOrder[2] = RESOLVER_BCAST;
			}
		} else {
			int[] tmp = new int[3];
			StringTokenizer st = new StringTokenizer( ro, "," );
			int i = 0;
			while( st.hasMoreTokens() ) {
				String s = st.nextToken().trim();
				if( s.equalsIgnoreCase( "LMHOSTS" )) {
					tmp[i++] = RESOLVER_LMHOSTS;
				} else if( s.equalsIgnoreCase( "WINS" )) {
					if( nbns == null ) {
						Log.println( Log.WARNINGS, "netbios name service warning",
								" resolveOrder specifies WINS however the " +
								"jcifs.netbios.wins property has not been set" );
						continue;
					}
					tmp[i++] = RESOLVER_WINS;
				} else if( s.equalsIgnoreCase( "BCAST" )) {
					tmp[i++] = RESOLVER_BCAST;
				} else if( s.equalsIgnoreCase( "DNS" )) {
					; // skip
				} else {
					Log.println( Log.WARNINGS, "netbios name service warning",
								"unknown resolver method: " + s );
				}
			}
			resolveOrder = new int[i];
			System.arraycopy( tmp, 0, resolveOrder, 0, i );
		}
    }

	int getNextNameTrnId() {
		if(( ++nextNameTrnId & 0xFFFF ) == 0 ) {
			nextNameTrnId = 1;
		}
		return nextNameTrnId;
	}
    void ensureOpen( int timeout ) throws IOException {
		closeTimeout = 0;
		if( soTimeout != 0 ) {
			closeTimeout = Math.max( soTimeout, timeout );
		}
		// If socket is still good, the new closeTimeout will
		// be ignored; see tryClose comment.
        if( socket == null ) {
        	socket = new DatagramSocket( port, laddr );
        	thread = new Thread( this, "JCIFS-NameServiceClient" );
			thread.setDaemon( true );
        	thread.start();
		}
    }
    void tryClose() {
		synchronized( socketLock ) {

			/* Yes, there is the potential to drop packets
			 * because we might close the socket during a
			 * request. However the chances are slim and the
			 * retry code should ensure the overall request
			 * is serviced. The alternative complicates things
			 * more than I think is worth it.
			 */

        	if( socket != null ) {
        	   	socket.close();
        		socket = null;
        	}
        	thread = null;
			responseTable.clear();
		}
    }
    public void run() {
		int nameTrnId;
		NameServicePacket response;

        while( thread == Thread.currentThread() ) {
			in.setLength( rcv_buf_size );
        	try {
        		socket.setSoTimeout( closeTimeout );
            	socket.receive( in );
        	} catch( IOException ioe ) {
				tryClose();
				break;
        	}

			Log.println( Log.DEBUGGING, "nbt name service debugging",
												" new data read from socket" );

			nameTrnId = NameServicePacket.readNameTrnId( rcv_buf, 0 );
			response = (NameServicePacket)responseTable.get( new Integer( nameTrnId ));
			if( response == null || response.received ) {
				continue;
			}
			synchronized( response ) {
				response.readWireFormat( rcv_buf, 0 );
				response.received = true;

				Log.printPacketData( "nbt name service packet receviced", response );
				Log.printHexDump( "datagram packet received from: " +
						in.getAddress().getHostAddress(), rcv_buf, 0, in.getLength() );

				response.notify();
			}
        }
    }
    void send( NameServicePacket request, NameServicePacket response,
											int timeout ) throws IOException {
		Integer nid = null;

		synchronized( response ) {
			try {
				synchronized( socketLock ) {
					request.nameTrnId = getNextNameTrnId();
					nid = new Integer( request.nameTrnId );

		        	out.setAddress( request.addr );
		        	out.setLength( request.writeWireFormat( snd_buf, 0 ));
					response.received = false;

					responseTable.put( nid, response );
		        	ensureOpen( timeout + 1000 );
		            socket.send( out );

					Log.printPacketData( "nbt name service packet sent", request );
					Log.printHexDump( "datagram packet sent to: " +
								out.getAddress().getHostAddress(), snd_buf, 0, out.getLength() );
				}

				response.wait( timeout );

			} catch( InterruptedException ie ) {
			} finally {
				responseTable.remove( nid );
			}
		}
    }

	NbtAddress getByName( Name name, InetAddress addr )
											throws UnknownHostException {
		int n;
		NameQueryRequest request = new NameQueryRequest( name );
		NameQueryResponse response = new NameQueryResponse();

		if( addr != null ) { /* UniAddress calls always use this
							  * because it specifies addr
							  */
			request.addr = addr; /* if addr ends with 255 flag it bcast */
			request.isBroadcast = (addr.getAddress()[3] == (byte)0xFF);

			n = retryCount;
			do {
				try {
					send( request, response, retryTimeout );
				} catch( IOException ioe ) {
					Log.printStackTrace( "nbt name service send:", ioe );
					throw new UnknownHostException( ioe.getMessage() );
				}

				if( response.received && response.resultCode == 0 ) {
					response.addrEntry.hostName.srcHashCode = addr.hashCode();
					return response.addrEntry;
				}
			} while( --n > 0 && request.isBroadcast );

			throw new UnknownHostException( name.toString() );
		}

		/* If a target address to query was not specified explicitly
		 * with the addr parameter we fall into this resolveOrder routine.
		 */

		for( int i = 0; i < resolveOrder.length; i++ ) {
			try {
				switch( resolveOrder[i] ) {
					case RESOLVER_LMHOSTS:
						NbtAddress ans = Lmhosts.getByName( name );
						if( ans != null ) {
							ans.hostName.srcHashCode = 0; // just has to be different
														  // from other methods
							return ans;
						}
						break;
					case RESOLVER_WINS:
					case RESOLVER_BCAST:
						if( resolveOrder[i] == RESOLVER_WINS &&
								name.name != NbtAddress.MASTER_BROWSER_NAME &&
								name.hexCode != 0x1d ) {
							request.addr = nbns;
							request.isBroadcast = false;
						} else {
							request.addr = baddr;
							request.isBroadcast = true;
						}

						n = retryCount;
						while( n-- > 0 ) {
							try {
								send( request, response, retryTimeout );
							} catch( IOException ioe ) {
								Log.printStackTrace( "nbt name service send:", ioe );
								throw new UnknownHostException( ioe.getMessage() );
							}
							if( response.received && response.resultCode == 0 ) {

/* Before we return, in anticipation of this address being cached we must
 * augment the addresses name's hashCode to distinguish those resolved by
 * Lmhosts, WINS, or BCAST. Otherwise a failed query from say WINS would
 * get pulled out of the cache for a BCAST on the same name.
 */
								response.addrEntry.hostName.srcHashCode =
														request.addr.hashCode();
								return response.addrEntry;
							} else if( resolveOrder[i] == RESOLVER_WINS ) {
								/* If WINS reports negative, no point in retry
								 */
								break;
							}
						}
						break;
				}
			} catch( IOException ioe ) {
			}
		}
		throw new UnknownHostException( name.name );
	}
	NbtAddress[] getNodeStatus( NbtAddress addr ) throws UnknownHostException {
		int n, srcHashCode;
		NodeStatusRequest request;
		NodeStatusResponse response;

		response = new NodeStatusResponse( addr );
		request = new NodeStatusRequest(
							new Name( NbtAddress.ANY_HOSTS_NAME, 0x00, null));
		request.addr = addr.getInetAddress();

		n = retryCount;
		while( n-- > 0 ) {
			try {
				send( request, response, retryTimeout );
			} catch( IOException ioe ) {
				Log.printStackTrace( "nbt name service send:", ioe );
				throw new UnknownHostException( ioe.getMessage() );
			}
			if( response.received && response.resultCode == 0 ) {

		/* For name queries resolved by different sources (e.g. WINS,
		 * BCAST, Node Status) we need to augment the hashcode generated
		 * for the addresses hostname or failed lookups for one type will
		 * be cached and cause other types to fail even though they may
		 * not be the authority for the name. For example, if a WINS lookup
		 * for FOO fails and caches unknownAddress for FOO, a subsequent
		 * lookup for FOO using BCAST should not fail because of that
		 * name cached from WINS.
		 *
		 * So, here we apply the source addresses hashCode to each name to
		 * make them specific to who resolved the name.
		 */

				srcHashCode = request.addr.hashCode();
				for( int i = 0; i < response.addressArray.length; i++ ) {
					response.addressArray[i].hostName.srcHashCode = srcHashCode;
				}
				return response.addressArray;
			}
		}
		throw new UnknownHostException( addr.hostName.name );
	}
}
