/* jcifs smb client library in Java
 * Copyright (C) 2000  "Michael B. Allen" <jcifs at samba dot org>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jcifs.smb;

import java.util.Vector;
import java.util.Enumeration;
import java.net.InetAddress;
import java.net.UnknownHostException;
import jcifs.UniAddress;

/**
 * The class represents a user's session established with an SMB/CIFS
 * server. This class is used internally to the jCIFS library however
 * applications may wish to authenticate aribrary user credentials
 * with the <tt>logon</tt> method. It is noteworthy that jCIFS does not
 * support DCE/RPC at this time and therefore does not use the NETLOGON
 * procedure. Instead, it simply performs a "tree connect" to IPC$ using
 * the supplied credentials. This is only a subset of the NETLOGON procedure
 * but is achives the same effect.
 */

public final class SmbSession {

	public static byte[] getChallenge( UniAddress dc ) throws SmbException, UnknownHostException {
		SmbTransport trans = SmbTransport.getSmbTransport( dc, 0 );
		trans.negotiate();
		return trans.server.encryptionKey;
	}
/**
 * Authenticate arbitrary credentials represented by the
 * <tt>NtlmPasswordAuthentication</tt> object against the domain controller
 * specified by the <tt>UniAddress</tt> parameter. If the credentials are
 * not accepted, an <tt>SmbAuthException</tt> will be thrown. If an error
 * occurs an <tt>SmbException</tt> will be thrown. If the credentials are
 * valid, the method will return without throwing an exception. See the
 * last <a href="../../../../FAQ.html">FAQ</a> question.
 */
	public static void logon( UniAddress dc,
						NtlmPasswordAuthentication auth ) throws SmbException {
		SmbTransport.getSmbTransport( dc, 0 ).getSmbSession( auth ).getSmbTree( "IPC$", null ).treeConnect( null, null );
	}

	int uid;
	NtlmPasswordAuthentication auth;
	SmbTransport transport;
	Vector trees;
	boolean sessionSetup;

	SmbSession( SmbTransport transport, NtlmPasswordAuthentication auth ) {
		this.transport = transport;
		this.auth = auth;
		trees = new Vector();
	}

	synchronized SmbTree getSmbTree( String share, String service ) {
		SmbTree t;

		if( share == null ) {
			share = "IPC$";
		}
		for( Enumeration e = trees.elements(); e.hasMoreElements(); ) {
			t = (SmbTree)e.nextElement();
			if( t.matches( share, service )) {
				return t;
			}
		}
		t = new SmbTree( this, share, service );
		trees.addElement( t );
		return t;
	}
	boolean matches( NtlmPasswordAuthentication auth ) {
		return this.auth == auth || this.auth.equals( auth );
	}
	void sendTransaction( SmbComTransaction request,
							SmbComTransactionResponse response ) throws SmbException {
		// transactions are not batchable
		sessionSetup( null, null );
		request.uid = uid;
		request.auth = auth;
		transport.sendTransaction( request, response );
	}
	void send( ServerMessageBlock request,
							ServerMessageBlock response ) throws SmbException {
		if( response != null ) {
			response.received = false;
		}
		sessionSetup( request, response );
		if( response != null && response.received ) {
			return;
		}
		request.uid = uid;
		request.auth = auth;
		transport.send( request, response );
	}
	void sessionSetup( ServerMessageBlock andx,
							ServerMessageBlock andxResponse ) throws SmbException {
synchronized( transport ) {
		if( sessionSetup ) {
			return;
		}

		transport.negotiate();

		if( transport.useSigning && transport.macSigningKey == null &&
				NtlmPasswordAuthentication.NULL != auth &&
				NtlmPasswordAuthentication.NULL.equals( auth ) == false ) {
			//	NtlmPasswordAuthentication.GUEST.equals( auth ) == false ) {
                    /* The first SMB_COM_SESSION_SETUP_ANX with creds other than
                     * null or guest generates the signing key */
			transport.initSigning( auth );
		}

		Log.println( Log.WARNINGS, "smb session setup warning",
					" requesting session with accountName=" + auth.username +
					",primaryDomain=" + auth.domain );

		/*
		 * Session Setup And X Request / Response
		 */

		SmbComSessionSetupAndX request = new SmbComSessionSetupAndX( this, andx );
		SmbComSessionSetupAndXResponse response = new SmbComSessionSetupAndXResponse( andxResponse );
		request.auth = auth;
		transport.send( request, response );

		uid = response.uid;
		sessionSetup = true;
}
	}
	void logoff( boolean inError ) {
synchronized( transport ) {
		if( sessionSetup == false ) {
			return;
		}

		for( Enumeration e = trees.elements(); e.hasMoreElements(); ) {
			SmbTree t = (SmbTree)e.nextElement();
			t.treeDisconnect( inError );
		}

		if( transport.server.security == ServerMessageBlock.SECURITY_SHARE ) {
			return;
		}

		if( !inError ) {

			/*
			 * Logoff And X Request / Response
			 */

			try {
				SmbComLogoffAndX request = new SmbComLogoffAndX( null );
				request.uid = uid;
				transport.send( request, null );
			} catch( SmbException se ) {
			}
		}
		sessionSetup = false;
}
	}
	public String toString() {
		return "SmbSession[accountName=" + auth.username +
				",primaryDomain=" + auth.domain +
				",uid=" + uid +
				",sessionSetup=" + sessionSetup + "]";
	}
}
