/* jcifs smb client library in Java
 * Copyright (C) 2000  "Michael B. Allen" <jcifs at samba dot org>
 *					"Eric Glass" <jcifs at samba dot org>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jcifs.smb;

import jcifs.netbios.NbtSocket;
import jcifs.netbios.NbtException;
import jcifs.netbios.NbtAddress;
import jcifs.UniAddress;
import jcifs.Config;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PushbackInputStream;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.net.InetAddress;
import java.net.SocketException;
import java.net.UnknownHostException;

import java.security.MessageDigest;

import java.util.Arrays;
import java.util.Vector;
import java.util.LinkedList;
import java.util.ListIterator;
import java.util.Enumeration;
import java.util.Hashtable;

class SmbTransport implements Runnable {

	private static final byte[] LMV2_CROSSDOMAIN_KEY = new byte[16];

	private static final int DEFAULT_MAX_MPX_COUNT = 10;
	private static final int DEFAULT_RESPONSE_TIMEOUT = 10000;
	private static final int DEFAULT_SO_TIMEOUT	= 15000;
	private static final int PUSHBACK_BUF_SIZE	 = 64;
	private static final int DEFAULT_RCV_BUF_SIZE  = 60416;
	private static final int DEFAULT_SND_BUF_SIZE  = 5000;
	private static final int DEFAULT_SSN_LIMIT	 = 250;

	private static final int LM_COMPATIBILITY =
			Config.getInt("jcifs.smb.lmCompatibility", 0);

	static final int MID_OFFSET	  = 30;
	static final int HEADER_LENGTH   = 32;
	static final int FLAGS_OFFSET	= 9;
	static final int FLAGS_RESPONSE  = 0x80;

	static final int ST_GROUND = 0;
	static final int ST_NEGOTIATING = 1;
	static final int ST_NEED_REFER = 2;
	static final int ST_GETTING_REFER = 3;
	static final int ST_NEGOTIATED = 4;

	private NbtSocket socket; // should become UniSocket?
	private InputStream in;
	private OutputStream out;
	private int localPort, soTimeout, responseTimeout;
	private long closeTime;
	private InetAddress localAddr;
	private Hashtable responseTable;
	private Thread thread;
	private Object outLock;

	private MessageDigest signingDigest;

	private static Vector connections = new Vector();
	private static MpxControl mpxCtrl = new MpxControl();

private static byte[] snd_buf = new byte[0xFFFF];
private static byte[] rcv_buf = new byte[0xFFFF];

	LinkedList sessions;
	boolean negotiated, useUnicode, useSigning;

	UniAddress address;

	byte[] macSigningKey;
	int signSequence;

	int port, rcv_buf_size, snd_buf_size;

	int negotiatedDialectIndex;
	int negotiatedMaxMpxCount;
	int negotiatedMaxBufferSize;
	int negotiatedCapabilities;

	int ssnLimit = Config.getInt( "jcifs.smb.client.ssnLimit", DEFAULT_SSN_LIMIT );

	int state;

	ClientProperties client = new ClientProperties();
	ServerProperties server = new ServerProperties();

	class ClientProperties {
		int maxMpxCount = Config.getInt( "jcifs.smb.client.maxMpxCount",
													DEFAULT_MAX_MPX_COUNT );
		int maxBufferSize = Config.getInt( "jcifs.smb.client.snd_buf_size",
													DEFAULT_SND_BUF_SIZE );
		int sessionKey = 0x00000000;
		/* NT 4 Workstation client capabilities 0x00D4
		 * we don't do NT Status Codes or Level II oplocks
		 * but we could do raw and mpx
		 */
		int flags2 = Config.getInt( "jcifs.smb.client.flags2",
				ServerMessageBlock.FLAGS2_LONG_FILENAMES |
				ServerMessageBlock.FLAGS2_EXTENDED_ATTRIBUTES |
				ServerMessageBlock.FLAGS2_UNICODE );
		int capabilities = Config.getInt( "jcifs.smb.client.capabilities",
 				ServerMessageBlock.CAP_UNICODE |
				ServerMessageBlock.CAP_NT_SMBS);
		String nativeOs = Config.getProperty( "jcifs.smb.client.nativeOs",
				System.getProperty( "os.name" ));
		String nativeLanMan = Config.getProperty( "jcifs.smb.client.nativeLanMan", "jCIFS" );
		int vcNumber = 1;
	}
	class ServerProperties {
		String tconHostName;

		byte flags;
		int flags2;
		int maxMpxCount;
		int maxBufferSize;
		int sessionKey;
		// NT 4 Workstation is 0x43FD
		int capabilities;
		String oemDomainName;
		String primaryDomain;
		String nativeOs;
		String nativeLanMan;

		int securityMode;
		int security;
		boolean encryptedPasswords;
		boolean signaturesEnabled;
		boolean signaturesRequired;
		int maxNumberVcs;
		int maxRawSize;
		long serverTime;
		int serverTimeZone;
		int encryptionKeyLength;
		byte[] encryptionKey;
	}

	static synchronized SmbTransport getSmbTransport( UniAddress address, int port ) {
		return getSmbTransport( address, port,
							Config.getInetAddress( "jcifs.smb.client.laddr", null ),
							Config.getInt( "jcifs.smb.client.lport", 0 ));
	}
	static synchronized SmbTransport getSmbTransport( UniAddress address, int port,
									InetAddress localAddr, int localPort ) {
		SmbTransport conn;
		for( Enumeration e = connections.elements(); e.hasMoreElements(); ) {
			conn = (SmbTransport)e.nextElement();
			if( conn.matches( address, port, localAddr, localPort )) {
				return conn;
			}
		}

		conn = new SmbTransport( address, port, localAddr, localPort );
		connections.addElement( conn );
		return conn;
	}

	SmbTransport( UniAddress address, int port, InetAddress localAddr, int localPort ) {
		this.address = address;
		this.port = port;
		this.localAddr = localAddr;
		this.localPort = localPort;

		useUnicode = Config.getBoolean( "jcifs.smb.client.useUnicode", true );
		if( useUnicode == false ) {
			client.flags2 &= 0xFFFF ^ ServerMessageBlock.FLAGS2_UNICODE;
			client.capabilities &= 0xFFFF ^ ServerMessageBlock.CAP_UNICODE;
		}
		if( Config.getBoolean( "jcifs.smb.client.useNTSmbs", false ) == true ) {
			client.capabilities |= ServerMessageBlock.CAP_NT_SMBS;
		}
		useSigning = Config.getBoolean("jcifs.smb.client.signingPreferred", false);
		if( useSigning ) {
			client.flags2 |= ServerMessageBlock.FLAGS2_SECURITY_SIGNATURES;
		}

		soTimeout = Config.getInt( "jcifs.smb.client.soTimeout", DEFAULT_SO_TIMEOUT );
		responseTimeout = Config.getInt( "jcifs.smb.client.responseTimeout", DEFAULT_RESPONSE_TIMEOUT );
		rcv_buf_size = Config.getInt( "jcifs.smb.client.rcv_buf_size", DEFAULT_RCV_BUF_SIZE );
		snd_buf_size = client.maxBufferSize;
		sessions = new LinkedList();
		responseTable = new Hashtable();
		outLock = new Object();
		state = ST_GROUND;
	}

	synchronized SmbSession getSmbSession() {
		return getSmbSession( new NtlmPasswordAuthentication( null, null, null ));
	}
	synchronized SmbSession getSmbSession( NtlmPasswordAuthentication auth ) {
		SmbSession ssn;

		ListIterator iter = sessions.listIterator();
		while( iter.hasNext() ) {
			ssn = (SmbSession)iter.next();
			if( ssn.matches( auth )) {
				ssn.auth = auth;
				return ssn;
			}
		}
		ssn = new SmbSession( this, auth );
		sessions.add( ssn );

		if( sessions.size() > ssnLimit ) {
			int nclose = sessions.size() / 10;
			SmbSession s;

			while( nclose-- > 0 ) {
				s = (SmbSession)sessions.removeFirst();
				s.logoff( false );
			}
		}

		return ssn;
	}
	boolean matches( UniAddress address, int port, InetAddress localAddr, int localPort ) {
		InetAddress defaultLocal = null;
		try {
			defaultLocal = InetAddress.getLocalHost();
		} catch( UnknownHostException uhe ) {
		}
		int p1 = ( port == 0 || port == 139 ) ? 0 : port;
		int p2 = ( this.port == 0 || this.port == 139 ) ? 0 : this.port;
		InetAddress la1 = localAddr == null ? defaultLocal : localAddr;
		InetAddress la2 = this.localAddr == null ? defaultLocal : this.localAddr;
		return address.equals( this.address ) &&
					p1 == p2 &&
					la1.equals( la2 ) &&
					localPort == this.localPort;
	}
	boolean hasCapability( int cap ) throws SmbException {
		if (state == ST_GROUND) {
			negotiate();
		}
		return (negotiatedCapabilities & cap) == cap;
	}
	void ensureOpen() throws IOException {
		if( socket == null ) {
			Object obj;
			NbtAddress naddr;
			String calledName;

			/* Hack to convert InetAddress to NbtAddress until we properly
			 * abstract NetBIOSless transport with some kind of "UniSocket".
			 */
			obj = address.getAddress();
			if( obj instanceof NbtAddress ) {
				naddr = (NbtAddress)obj;
			} else {
				try {
					naddr = NbtAddress.getByName( ((InetAddress)obj).getHostAddress() );
				} catch( UnknownHostException uhe ) {
					naddr = null; // never happen
				}
			}

			calledName = address.firstCalledName();
			do {
				try {
					socket = new NbtSocket( naddr, calledName, port, localAddr, localPort );
					break;
				} catch( NbtException ne ) {
					if( ne.errorClass == NbtException.ERR_SSN_SRVC &&
								( ne.errorCode == NbtException.CALLED_NOT_PRESENT ||
								ne.errorCode == NbtException.NOT_LISTENING_CALLED )) {
						Log.println( Log.WARNINGS, "smb warning", ne.getMessage() );
					} else {
						throw ne;
					}
				}
			} while(( calledName = address.nextCalledName()) != null );

			if( calledName == null ) {
				throw new IOException( "Failed to establish session with " + address );
			}

			/* Save the calledName for using on SMB_COM_TREE_CONNECT
			 */
			if( calledName == NbtAddress.SMBSERVER_NAME ) {
				server.tconHostName = address.getHostAddress();
			} else {
				server.tconHostName = calledName;
			}

			if( Config.getBoolean( "jcifs.smb.client.tcpNoDelay", false )) {
				socket.setTcpNoDelay( true );
			}
			in = new PushbackInputStream( socket.getInputStream(), PUSHBACK_BUF_SIZE );
			out = socket.getOutputStream();
			thread = new Thread( this, "JCIFS-SmbTransport" );
			thread.setDaemon( true );
			thread.start();
		}
	}
	void tryClose( boolean inError ) {
		SmbSession ssn;

		if( socket == null ) {
			inError = true;
		}

		ListIterator iter = sessions.listIterator();
		while( iter.hasNext() ) {
			ssn = (SmbSession)iter.next();
			ssn.logoff( inError );
		}
		if( socket != null ) {
			try {
				socket.close();
			} catch( IOException ioe ) {
			}
		}
		in = null;
		out = null;
		socket = null;
		thread = null;
		responseTable.clear();
		state = ST_GROUND;
	}
	public void run() {
		int mid, l, i, n, m;
		ServerMessageBlock response;
		int magic[] = { 0xFF, 'S', 'M', 'B' };

		while( thread == Thread.currentThread() ) {
			try {
				socket.setSoTimeout( soTimeout );

				m = 0;
				while( m < 4 ) {
					if(( i = in.read() ) < 0 ) {
						return;
					}
					if(( i & 0xFF ) == magic[m] ) {
						m++;
					} else if(( i & 0xFF ) == 0xFF ) {
						m = 1;
					} else {
						m = 0;
					}
				}

synchronized( rcv_buf ) {
				rcv_buf[0] = (byte)0xFF;
				rcv_buf[1] = (byte)'S';
				rcv_buf[2] = (byte)'M';
				rcv_buf[3] = (byte)'B';

				if( in.read( rcv_buf, 4, HEADER_LENGTH - 4 ) != ( HEADER_LENGTH - 4 )) {
					/* Read on a netbios SocketInputStream does not
					 * return until len bytes have been read or the stream is
					 * closed. This must be the latter case.
					 */
					break;
				}
				((PushbackInputStream)in).unread( rcv_buf, 0, HEADER_LENGTH );
				if( rcv_buf[0] != (byte)0xFF ||
								rcv_buf[1] != (byte)'S' ||
								rcv_buf[2] != (byte)'M' ||
								rcv_buf[3] != (byte)'B' ) {
					Log.println( Log.WARNINGS, "smb warning",
												"bad smb header, purging session message" );
					in.skip( in.available() );
					continue;
				}
				if(( rcv_buf[FLAGS_OFFSET] & FLAGS_RESPONSE ) == FLAGS_RESPONSE ) {
					mid = ServerMessageBlock.readInt2( rcv_buf, MID_OFFSET );

					response = (ServerMessageBlock)responseTable.get( new Integer( mid ));
					if( response == null ) {
						Log.println( Log.WARNINGS, "smb warning",
								" no handler for mid=" + mid + ", purging session message" );
						in.skip( in.available() );
						continue;
					}
					synchronized( response ) {
						response.useUnicode = useUnicode;
						Log.println( Log.DEBUGGING, "smb transport warning",
														" new data read from socket" );

						if( response instanceof SmbComTransactionResponse ) {
							Enumeration e = (Enumeration)response;
							if( e.hasMoreElements() ) {
								e.nextElement();
							} else {
								Log.println( Log.WARNINGS, "smb warning",
											"more responses to transaction than expected" );
								continue;
							}
							response.readWireFormat( in, rcv_buf, 0 );
							response.received = true;
							Log.printMessageData( "smb received", response );

							if( response.errorCode != 0 || e.hasMoreElements() == false ) {
								((SmbComTransactionResponse)response).hasMore = false;
								if( useSigning ) {
									response.verifyFailed = verify(rcv_buf, 0, response.length, response.verifySequence);
								}
								response.notify();
							} else {
								ensureOpen();
							}
						} else {
							response.readWireFormat( in, rcv_buf, 0 );
							response.received = true;

							Log.printMessageData( "smb received", response );
							ServerMessageBlock smb = response;
							while( smb instanceof AndXServerMessageBlock &&
									( smb = ((AndXServerMessageBlock)smb).andx ) != null ) {
								Log.printMessageData( "smb andx data", smb );
							}
							Log.printHexDump( "smb received", rcv_buf, 0, response.length );
							if( useSigning ) {
								response.verifyFailed = verify(rcv_buf, 0, response.length, response.verifySequence);
							}

							response.notify();
						}
					}
				} else {
					// it's a request(break oplock)
				}
}
			} catch( InterruptedIOException iioe ) {
				if( responseTable.size() == 0 ) {
					tryClose( false );
				} else {
					Log.println( Log.WARNINGS, "smb warning",
								" soTimeout has occured but there are " +
								responseTable.size() + " pending requests" );
				}
			} catch( IOException ioe ) {
				synchronized( this ) {
					tryClose( true );
				}
				Log.printStackTrace( "exception reading from socket input: " + address, ioe );
			}
		}
	}

	void initSigning(NtlmPasswordAuthentication auth) throws SmbException {
		if( auth.hashesExternal ) {
			if( server.signaturesRequired ) {
				throw new SmbException( SmbException.ERRCLI, SmbException.ERRioe,
						"Signing is required by the server but passwords are external." );
			}
			useSigning = false;
			return;
		}

		signSequence = 0;

		switch (LM_COMPATIBILITY) {
		case 0:
		case 1:
		case 2:
			macSigningKey = new byte[40];
			try {
				auth.getUserSessionKey(server.encryptionKey, macSigningKey, 0);
				System.arraycopy(auth.getUnicodeHash(server.encryptionKey),
						0, macSigningKey, 16, 24);
			} catch (Exception ex) {
				Log.printStackTrace("Unable to calculate MAC signing key.", ex);
				macSigningKey = null;
			}
			break;
		case 3:
		case 4:
		case 5:
			macSigningKey = new byte[16];
			try {
				auth.getUserSessionKey(server.encryptionKey, macSigningKey, 0);
			} catch (Exception ex) {
				Log.printStackTrace("Unable to calculate MAC signing key.", ex);
				macSigningKey = null;
			}
			break;
		default:
			macSigningKey = new byte[40];
			try {
				auth.getUserSessionKey(server.encryptionKey, macSigningKey, 0);
				System.arraycopy(auth.getUnicodeHash(server.encryptionKey),
						0, macSigningKey, 16, 24);
			} catch (Exception ex) {
				Log.printStackTrace("Unable to calculate MAC signing key.", ex);
				macSigningKey = null;
			}
			break;
		}
	}

	/**
	 * Performs MAC signing of the SMB.  This is done as follows.
	 * The signature field of the SMB is overwritted with the sequence number;
	 * The MD5 digest of the MAC signing key + the entire SMB is taken;
	 * The first 8 bytes of this are placed in the signature field.
	 *
	 * @param data The data.
	 * @param offset The starting offset at which the SMB header begins.
	 * @param length The length of the SMB data starting at offset. 
	 */
	private void sign(byte[] data, int offset, int length) {
		if (macSigningKey == null) return;
		try {
			signingDigest.update(macSigningKey);
			int index = offset + ServerMessageBlock.SIGNATURE_OFFSET;
			for (int i = 0; i < 8; i++) data[index + i] = 0;
			ServerMessageBlock.writeInt4(signSequence, data, index);
			signingDigest.update(data, offset, length);
			System.arraycopy(signingDigest.digest(), 0, data, index, 8);
		} catch (Exception ex) {
			Log.printStackTrace("Error signing SMB.", ex);
		} finally {
			signSequence += 2;
		}
	} 

	/**
	 * Performs MAC signature verification.  This calculates the signature
	 * of the SMB and compares it to the signature field on the SMB itself.
	 *
	 * @param data The data.
	 * @param offset The starting offset at which the SMB header begins.
	 * @param length The length of the SMB data starting at offset. 
	 */
	private boolean verify(byte[] data, int offset, int length, int verifySequence) throws IOException {
		if (macSigningKey == null) return false;
		signingDigest.update(macSigningKey);
		int index = offset;
		signingDigest.update(data, index, ServerMessageBlock.SIGNATURE_OFFSET); 
		index += ServerMessageBlock.SIGNATURE_OFFSET;
		byte[] sequence = new byte[8]; 
		ServerMessageBlock.writeInt4(verifySequence, sequence, 0); 
		signingDigest.update(sequence); 
		index += 8;
		signingDigest.update(data, index, length - ServerMessageBlock.SIGNATURE_OFFSET - 8);
		byte[] signature = signingDigest.digest();
		for (int i = 0; i < 8; i++) {
			if (signature[i] != data[offset + ServerMessageBlock.SIGNATURE_OFFSET + i]) {
											// can this just use == ?
				if (Arrays.equals(LMV2_CROSSDOMAIN_KEY, macSigningKey)) return true;
				signingDigest.update(LMV2_CROSSDOMAIN_KEY);
				index = offset;
				signingDigest.update(data, index, ServerMessageBlock.SIGNATURE_OFFSET);
				index += ServerMessageBlock.SIGNATURE_OFFSET;
				ServerMessageBlock.writeInt4(verifySequence, sequence, 0);
				signingDigest.update(sequence);
				index += 8;
				signingDigest.update(data, index, length - ServerMessageBlock.SIGNATURE_OFFSET - 8);
				signature = signingDigest.digest();
				for (i = 0; i < 8; i++) {
					if (signature[i] != data[offset + ServerMessageBlock.SIGNATURE_OFFSET + i]) {
						return true;
					}
				}
				// verified using LMv2 cross-domain MAC key; start using that.
				macSigningKey = LMV2_CROSSDOMAIN_KEY;
				break;
			}
		}

		return false;
	} 

	synchronized DfsReferral getDfsReferral( NtlmPasswordAuthentication auth, String path ) throws SmbException {
		String subpath, node, host;
		DfsReferral dr = new DfsReferral();
		int p, n, i, s;
		UniAddress addr;

		SmbTree ipc = getSmbSession( auth ).getSmbTree( "IPC$", null );
		Trans2GetDfsReferralResponse resp = new Trans2GetDfsReferralResponse();
		ipc.sendTransaction( new Trans2GetDfsReferral( path ), resp );

		subpath = path.substring( 0, resp.pathConsumed );
		node = resp.referral.node;
		if( subpath.charAt( 0 ) != '\\' ||
				(i = subpath.indexOf( '\\', 1 )) < 2 ||
				(p = subpath.indexOf( '\\', i + 1 )) < (i + 2) ||
				node.charAt( 0 ) != '\\' ||
				(s = node.indexOf( '\\', 1 )) < 2) {
			throw new SmbException( SmbException.ERRCLI, SmbException.ERRdfs, "Invalid DFS path: " + path );
		}
		if ((n = node.indexOf( '\\', s + 1 )) == -1) {
			n = node.length();
		}

		dr.path = subpath.substring( p );
		dr.node = node.substring( 0, n );
		dr.nodepath = node.substring( n );

		host = node.substring( 1, s );
		try {
			addr = UniAddress.getByName( host );
		} catch( UnknownHostException uhe ) {
			throw new SmbException( SmbException.ERRCLI, SmbException.ERRdfs, "unknown host: " + host );
		}
		SmbTransport trans = getSmbTransport( addr, 0 );
		dr.tree = trans.getSmbSession( auth ).getSmbTree( node.substring( s + 1, n ), null );

		return dr;
	}
	void send( ServerMessageBlock request,
							ServerMessageBlock response ) throws SmbException {
		Integer mid = null;

		if (state == ST_GROUND) {
			negotiate();
		}

		request.flags2 |= client.flags2;
		request.mid = aquireMid();
		request.useUnicode = useUnicode;

		if( response == null ) {
			try {
				synchronized( outLock ) {
					ensureOpen();
synchronized( snd_buf ) {
					int length = request.writeWireFormat(snd_buf, 0);
					if( useSigning ) {
						sign(snd_buf, 0, length);
					}
					out.write(snd_buf, 0, length);
					out.flush();

					Log.printMessageData( "smb sent", request );
					ServerMessageBlock smb = request;
					while( smb instanceof AndXServerMessageBlock &&
									( smb = ((AndXServerMessageBlock)smb).andx ) != null ) {
						Log.printMessageData( "smb andx data", smb );
					}
					Log.printHexDump( "smb sent", snd_buf, 0, request.length );
}
				}
			} catch( IOException ioe ) {
				tryClose( true );
				throw new SmbException( SmbException.ERRCLI, SmbException.ERRioe, ioe.getMessage() );
			} finally {
				releaseMid( request.mid );
			}

			return;
		}

		// now for the normal case where response is not null

		try {
			synchronized( response ) {

				response.received = false;
				mid = new Integer( request.mid );
				responseTable.put( mid, response );

				synchronized( outLock ) {
					ensureOpen();
synchronized( snd_buf ) {
					int length = request.writeWireFormat(snd_buf, 0);
					if( useSigning ) {
						response.verifySequence = signSequence + 1;
						sign(snd_buf, 0, length);
					}
					out.write(snd_buf, 0, length);
					out.flush();

					Log.printMessageData( "smb sent", request );
					ServerMessageBlock smb = request;
					while( smb instanceof AndXServerMessageBlock &&
									( smb = ((AndXServerMessageBlock)smb).andx ) != null ) {
						Log.printMessageData( "smb andx data", smb );
					}
					Log.printHexDump( "smb sent", snd_buf, 0, request.length );
}
				}

				// default it 1 so that 0 can be used as forever
				response.wait( response.responseTimeout == 1 ?
												responseTimeout : response.responseTimeout );
			}
		} catch( InterruptedException ie ) {
			tryClose( true );
		} catch( IOException ioe ) {
			tryClose( true );
			throw new SmbException( SmbException.ERRCLI, SmbException.ERRioe, ioe.getMessage() );
		} finally {
			responseTable.remove( mid );
			releaseMid( request.mid );
		}

		if( response.received == false ) {
			tryClose( true );
			throw new SmbException( SmbException.ERRCLI,
									SmbException.ERRserverTimeout,
									address );
		} else if( response.verifyFailed ) {
			tryClose( true );
			throw new SmbException( SmbException.ERRCLI,
									SmbException.ERRioe,
									"Unverifiable signature." );
		}
		switch( response.errorCode & 0xFF ) {
			case SmbException.SUCCESS:
				break;
			case SmbException.ERRDOS:
				switch(( response.errorCode >> 16 ) & 0xFFFF ) {
					case SmbException.ERRnoaccess:
						throw new SmbAuthException( response.errorCode );
				}
				throw new SmbException( response.errorCode );
			case SmbException.ERRSRV:
				switch(( response.errorCode >> 16 ) & 0xFFFF ) {
					case SmbException.ERRreserved:
						if( request.auth == null ) {
							throw new SmbException( response.errorCode );
						}
						throw getDfsReferral( request.auth, request.path );
					case SmbException.ERRbadpw:
					case SmbException.ERRaccess:
					case SmbException.ERRaccountExpired:
					case SmbException.ERRbadClient:
					case SmbException.ERRbadLogonTime:
					case SmbException.ERRpasswordExpired:
						throw new SmbAuthException( response.errorCode );
				}
			default:
				throw new SmbException( response.errorCode );
		}

	}
	void sendTransaction( SmbComTransaction request,
							SmbComTransactionResponse response ) throws SmbException {
		Integer mid = null;

		negotiate();

		request.flags2 |= client.flags2;
		request.mid = aquireMid();
		mid = new Integer( request.mid );
		request.useUnicode = useUnicode;
		request.maxBufferSize = negotiatedMaxBufferSize;
		response.received = false;
		response.hasMore = true;
		response.isPrimary = true;

		try {
			request.txn_buf = BufferCache.getBuffer();
			response.txn_buf = BufferCache.getBuffer();

			request.nextElement();
			if( request.hasMoreElements() ) {
				// multi-part request

				SmbComBlankResponse interimResponse = new SmbComBlankResponse();

				synchronized( interimResponse ) {

					responseTable.put( mid, interimResponse );

					synchronized( outLock ) {
						ensureOpen();
synchronized(snd_buf) {
						int length = request.writeWireFormat(snd_buf, 0);
						if( useSigning ) {
							response.verifySequence = signSequence + 1;
							sign(snd_buf, 0, length);
						}
						out.write(snd_buf, 0, length);
						out.flush();

						Log.printMessageData( "smb sent", request );
						Log.printHexDump( "smb sent", snd_buf, 0, request.length );
}
					}

					interimResponse.wait( responseTimeout );

					if( interimResponse.received == false ) {
						throw new SmbException( SmbException.ERRserverTimeout );
					}
					switch( interimResponse.errorCode & 0xFF ) {
						case SmbException.SUCCESS:
							break;
						case SmbException.ERRDOS:
							switch(( response.errorCode >> 16 ) & 0xFFFF ) {
								case SmbException.ERRnoaccess:
									throw new SmbAuthException( response.errorCode );
							}
							throw new SmbException( response.errorCode );
						case SmbException.ERRSRV:
							switch(( interimResponse.errorCode >> 16 ) & 0xFFFF ) {
								case SmbException.ERRreserved:
									if( request.auth == null ) {
										throw new SmbException( response.errorCode );
									}
									throw getDfsReferral( request.auth, request.path );
								case SmbException.ERRbadpw:
								case SmbException.ERRaccess:
								case SmbException.ERRaccountExpired:
								case SmbException.ERRbadClient:
								case SmbException.ERRbadLogonTime:
								case SmbException.ERRpasswordExpired:
									throw new SmbAuthException( interimResponse.errorCode );
							}
						default:
							throw new SmbException( interimResponse.errorCode );
					}
				}
				request.nextElement();
			}

			synchronized( response ) {
				responseTable.put( mid, response );
				do {
					synchronized( outLock ) {
						ensureOpen();
synchronized( snd_buf ) {
						int length = request.writeWireFormat(snd_buf, 0);
						if( useSigning ) {
							response.verifySequence = signSequence + 1;
							sign(snd_buf, 0, length);
						}
						out.write(snd_buf, 0, length);
						out.flush();

						Log.printMessageData( "smb sent", request );
						Log.printHexDump( "smb sent", snd_buf, 0, request.length );
}
					}
				} while( request.hasMoreElements() && request.nextElement() != null );

				do {
					// default it 1 so that 0 can be used as forever
					response.received = false;
					response.wait( response.responseTimeout == 1 ? responseTimeout : response.responseTimeout );
				} while( response.received && response.hasMoreElements() );
			}
		} catch( InterruptedException ie ) {
			tryClose( true );
		} catch( IOException ioe ) {
			tryClose( true );
			throw new SmbException( SmbException.ERRCLI, SmbException.ERRioe, ioe.getMessage() );
		} finally {
			responseTable.remove( mid );
			releaseMid( request.mid );
			BufferCache.releaseBuffer( request.txn_buf );
			BufferCache.releaseBuffer( response.txn_buf );
		}

		if( response.received == false ) {
			tryClose( true );
			throw new SmbException( SmbException.ERRCLI,
									SmbException.ERRserverTimeout,
									address );
		} else if( response.verifyFailed ) {
			tryClose( true );
			throw new SmbException( SmbException.ERRCLI,
									SmbException.ERRioe,
									"Unverifiable signature." );
		}
		switch( response.errorCode & 0xFF ) {
			case SmbException.SUCCESS:
				break;
			case SmbException.ERRDOS:
				switch(( response.errorCode >> 16 ) & 0xFFFF ) {
					case SmbException.ERRnoaccess:
						throw new SmbAuthException( response.errorCode );
				}
				throw new SmbException( response.errorCode );
			case SmbException.ERRSRV:
				switch(( response.errorCode >> 16 ) & 0xFFFF ) {
					case SmbException.ERRreserved:
						if( request.auth == null ) {
System.err.println( "req: " + request.getClass().getName() );
							throw new SmbException( response.errorCode );
						}
						throw getDfsReferral( request.auth, request.path );
					case SmbException.ERRbadpw:
					case SmbException.ERRaccess:
					case SmbException.ERRaccountExpired:
					case SmbException.ERRbadClient:
					case SmbException.ERRbadLogonTime:
					case SmbException.ERRpasswordExpired:
						throw new SmbAuthException( response.errorCode );
				}
			default:
				throw new SmbException( response.errorCode );
		}
	}
	synchronized void negotiate() throws SmbException {

		if( state >= ST_NEGOTIATING ) {
			return;
		}
		state = ST_NEGOTIATING;

		Log.println( Log.WARNINGS, "smb negotiation warning",
							" requesting negotiation with " + address );

		/*
		 * Negotiate Protocol Request / Response
		 */

		// reset MAC signing
		macSigningKey = null;

		SmbComNegotiateResponse response = new SmbComNegotiateResponse();
		send( new SmbComNegotiate(), response );

		if( response.dialectIndex > 10 ) {
			tryClose( true );
			throw new SmbException( SmbException.ERRCLI, SmbException.ERRbadDialect );
		}

		server.securityMode		= response.securityMode;
		server.security			= response.security;
		server.encryptedPasswords  = response.encryptedPasswords;
		server.signaturesEnabled   = response.signaturesEnabled;
		server.signaturesRequired  = response.signaturesRequired;
		if (server.signaturesRequired || (server.signaturesEnabled && useSigning)) {
			useSigning = true;
			client.flags2 |= ServerMessageBlock.FLAGS2_SECURITY_SIGNATURES;

			if (signingDigest == null) {
				try {
					signingDigest = MessageDigest.getInstance("MD5");
				} catch (Exception ex) {
					Log.printStackTrace("Unable to obtain MD5 digest.", ex);
					return;
				}
			}
		} else {
			useSigning = false;
			client.flags2 &= 0xFFFF ^ ServerMessageBlock.FLAGS2_SECURITY_SIGNATURES;
		}
		negotiatedDialectIndex	 = response.dialectIndex;;
		server.maxMpxCount		 = response.maxMpxCount;
		negotiatedMaxMpxCount	  = client.maxMpxCount < server.maxMpxCount ?
										client.maxMpxCount : server.maxMpxCount;
		mpxCtrl.maxMpxCount		= negotiatedMaxMpxCount < 1 ?
										1 : negotiatedMaxMpxCount;
		server.maxNumberVcs		= response.maxNumberVcs;
		server.maxBufferSize	   = response.maxBufferSize;
		negotiatedMaxBufferSize	= client.maxBufferSize < server.maxBufferSize ?
										client.maxBufferSize : server.maxBufferSize;
		server.maxRawSize		  = response.maxRawSize;
		server.sessionKey		  = response.sessionKey;
		server.capabilities		= response.capabilities;
		negotiatedCapabilities	 = client.capabilities & server.capabilities;
		if(( negotiatedCapabilities & ServerMessageBlock.CAP_UNICODE ) == 0 ) {
			// server doesn't want unicode
			if( Config.getBoolean( "jcifs.smb.client.useUnicode", false )) {
				// force unicode
				negotiatedCapabilities |= ServerMessageBlock.CAP_UNICODE;
			} else {
				// not explicitly set to true so flip unicode off as server requests
				useUnicode = false;
				client.flags2 &= 0xFFFF ^ ServerMessageBlock.FLAGS2_UNICODE;
			}
		}
		server.serverTime		  = response.serverTime;
		server.serverTimeZone	  = response.serverTimeZone;
		server.encryptionKeyLength = response.encryptionKeyLength;
		server.encryptionKey	   = response.encryptionKey;
		server.oemDomainName	   = response.oemDomainName;
	}
	public String toString() {
		String ret = "SmbTransport[address=" + address;
		if( socket == null ) {
			ret += ",port=,localAddr=,localPort=]";
		} else {
			ret += ",port=" + socket.getPort() +
			",localAddr=" + socket.getLocalAddress() +
			",localPort=" + socket.getLocalPort() + "]";
		}
		return ret;
	}

	static int aquireMid() throws SmbException {
		try {
			return mpxCtrl.aquireMid();
		} catch( InterruptedException ie ) {
			throw new SmbException( SmbException.ERRCLI, SmbException.ERRioe, ie.getMessage() );
		}
	}
	static void releaseMid( int mid ) {
		mpxCtrl.releaseMid( mid );
	}

	static class MpxControl {
		MpxListNode first, last;
		class MpxListNode {
			int i;  
			MpxListNode next;
			MpxListNode( int id ) {
				i = id;
				next = null;
			}
		}	   
		int nextMid, mpxCount, maxMpxCount;

		MpxControl() {
			nextMid = 0;
			mpxCount = 0;
			maxMpxCount = 1;
			first = last = null;
		}

		synchronized void clear() {
			nextMid = 0;
			mpxCount = 0;
			maxMpxCount = 1;
			first = last = null;
		}
		synchronized int aquireMid() throws InterruptedException {
			while( mpxCount >= maxMpxCount ) {
				wait();
			}
			if(( ++nextMid % 0xFFFF ) == 0 ) {
				nextMid = 1;
			}
			add( nextMid );
			mpxCount++;
			return nextMid;
		}
		synchronized void releaseMid( int i ) {
			remove( i );
			mpxCount--;
			notify();
		}
		synchronized boolean contains( int i ) {
			for( MpxListNode tmp = first; tmp != null; tmp = tmp.next ) {
				if( tmp.i == i ) {
					return true;
				}
			}
			return false;
		}	   
		synchronized void add( int i ) {
			if( contains( i )) {
				return;
			}
			if( first == null ) {
				first = last = new MpxListNode( i );
			} else {
				last = last.next = new MpxListNode( i );
			}
		}
		synchronized void remove( int i ) {
			if( first == null ) {
				return;
			} else if( first == last && first.i == i ) {
				first = last = null;
				return;
			}
			MpxListNode tmp, prev;
			for( tmp = prev = first; tmp != null; tmp = tmp.next ) {
				if( tmp.i == i ) {
					if( tmp == first ) {
						first = first.next;
					} else if( tmp == last ) {
						last = prev;
						last.next = null;
					} else {
						prev.next = tmp.next;
					}
					return;
				}
				prev = tmp;
			}
		}	   
	}
}
