/* jcifs smb client library in Java
 * Copyright (C) 2000  "Michael B. Allen" <jcifs at samba dot org>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jcifs.util;

import java.util.Properties;
import java.util.Enumeration;
import java.io.InputStream;
import java.io.PrintStream;
import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;

/**
 * This class uses a static {@link jcifs.util.PropertiesTree} to act
 * as a cental repository of configuration information. It cannot be
 * instantiated. Similar to <code>System</code> properties the namespace
 * is global therefore property names should be unique. Before use,
 * the <code>load</code> method should be called with the name of a
 * <code>PropertiesTree</code> file(or <code>null</code> indicating no
 * file) to initialize the <code>Config</code>. The <code>System</code>
 * properties will then populate the <code>Config</code> as well potentially
 * overwriting properties from the file. Thus properties provided on the
 * commandline with the <code>-Dproperty.name=value</code> VM parameter
 * will override properties from the configuration file.
 */

public class Config {

	// supress javadoc constructor summary by removing 'protected'
	Config() {}

	/**
	 * The static <code>PropertiesTree</code>.
	 */

	protected static PropertiesTree properties = new PropertiesTree();

	/**
	 * Load the <code>Config</code> with properties from the stream
	 * <code>in</code> which should be in <code>PropertiesTree</code>
	 * format. See <a href="../../../properties.html">More About
	 * PropertiesTree and it's File Format</code>.
	 */

	public static void load( InputStream in ) throws IOException {
		Properties sys;

		if( in != null ) {
			properties.load( in );
		}

		sys = System.getProperties();
		for( Enumeration e = sys.propertyNames(); e.hasMoreElements(); ) {
			String key = (String)e.nextElement();
			String val = (String)sys.getProperty( key );
			if( key != null && val != null ) {
				try {
					properties.put( key, val );
				} catch( IllegalArgumentException iae ) {
				}
			}
		}
	}

	/**
	 * List the properties in the <code>Code</code>.
	 */

	public static void list( PrintStream out ) throws IOException {
		properties.list( out );
	}

	/**
	 * Add a property. If the property already
	 * exists and it is a <code>PropertiesTree</code> an
	 * <code>IllegalArgumentException</code> will be thrown to indicate
	 * that an entire branch of the tree cannot be overwritten. If
	 * however the node is a leaf property, it will be overwritten
	 * and the old value returned as a <code>String</code>.
	 */

	public static String setProperty( String key, String value ) {
		return (String)properties.put( key, value );
	}

	/**
	 * Retrieve a property as an <code>Object</code>. Possible
	 * types are <code>String</code> if the property is a leaf node,
	 * or a <code>PropertiesTree</code> if key resolves to a parent node,
	 * or <code>null</code> otherwise.
	 */

	public static Object get( String key ) {
		return properties.get( key );
	}

	/**
	 * Retrieve a <code>String</code>. If the key cannot be found,
	 * the provided <code>def</code> default parameter will be returned.
	 */

	public static String getProperty( String key, String def ) {
		String result = null;
		Object obj = get( key );
		if( obj != null && obj instanceof String ) {
			result = (String)obj;
		}
		return obj == null ? def : result;
	}

	/**
	 * Retrieve a <code>String</code>. If the property is not found, <code>null</code> is returned.
	 */

	public static String getProperty( String key ) {
		String result = null;
		Object obj = get( key );
		if( obj != null && obj instanceof String ) {
			result = (String)obj;
		}
		return result;
	}

	/**
	 * Retrieve an <code>int</code>. If the key does not exist or
	 * cannot be converted to an <code>int</code>, the provided default
	 * argument will be returned.
	 */

	public static int getInt( String key, int def ) {
		Object obj = get( key );
		if( obj != null && obj instanceof String ) {
			try {
				def = Integer.parseInt( (String)obj );
			} catch( NumberFormatException nfe ) {
				Log.printStackTrace( "configuration warning", nfe );
			}
		}
		return def;
	}

	/**
	 * Retrieve an <code>int</code>. If the property is not found, <code>-1</code> is returned.
	 */

	public static int getInt( String key ) {
		int result = -1;
		Object obj = get( key );
		if( obj != null && obj instanceof String ) {
			try {
				result = Integer.parseInt( (String)obj );
			} catch( NumberFormatException nfe ) {
				Log.printStackTrace( "configuration warning", nfe );
			}
		}
		return result;
	}

	/**
	 * Retrieve a <code>long</code>. If the key does not exist or
	 * cannot be converted to a <code>long</code>, the provided default
	 * argument will be returned.
	 */

	public static long getLong( String key, long def ) {
		Object obj = get( key );
		if( obj != null && obj instanceof String ) {
			try {
				def = Long.parseLong( (String)obj );
			} catch( NumberFormatException nfe ) {
				Log.printStackTrace( "configuration warning", nfe );
			}
		}
		return def;
	}

	/** 
	 * Retrieve an <code>InetAddress</code>. If the address is not
	 * an IP address and cannot be resolved <code>null</code> will
	 * be returned.
	 */

	public static InetAddress getInetAddress( String key, InetAddress def ) {
		String addr = getProperty( key );
		if( addr != null ) {
			try {
				def = InetAddress.getByName( addr );
			} catch( UnknownHostException uhe ) {
			}
		}
		return def;
	}

	/**
	 * Retrieve a boolean value. If the property is not found, the value of <code>def</code> is returned.
	 */

	public static boolean getBoolean( String key, boolean def ) {
		String b = getProperty( key );
		if( b != null ) {
			def = b.toLowerCase().equals( "true" );
		}
		return def;
	}

	/**
	 * Print all properties through the {@link jcifs.util.Log} class with the mask <code>Log.WARNINGS</code>.
	 */

	public static void printProperties( String msg ) {
		Log.printProperties( Log.WARNINGS, msg, properties );
	}
}
